#include "propeller.hpp"
#include "propclock.hpp"
#include "messages.hpp"

#include <esp_log.h>
#include <freertos/FreeRTOS.h>
#include <freertos/task.h>
#include <driver/spi_master.h>
#include <driver/gpio.h>
#include <endian.h>

#include <cstring>
#include <array>


namespace {

// std::variant explicit deduction guide (not needed as of C++20)
template<class... Ts> struct overloaded : Ts... { using Ts::operator()...; };
template <class... Ts> overloaded(Ts...) -> overloaded<Ts...>;



const int MOSI = 13;
const int MISO = 12;
const int CLK = 14;
const auto CS = gpio_num_t(15);
const int SPI_SPEED = (1000 * 1000);

template <typename T>
datagram_t propeller_transaction(spi_device_handle_t spi, const T& output)
{
  datagram_t input;
  esp_err_t ret;
  spi_transaction_t t;
  std::memset(&t, 0, sizeof(t));       //Zero out the transaction
  t.length = 8 * 8;                     // datagram is 8 byte
  t.tx_buffer = output.data();
  t.rx_buffer = &input;
  gpio_set_level(CS, 0);
  ret = spi_device_transmit(spi, &t);  //Transmit!
  assert(ret == ESP_OK);            //Should have had no issues.
  gpio_set_level(CS, 1);
  input.clock_counter = be32toh(input.clock_counter);
  input.peripherals = be32toh(input.peripherals);
  return input;
}

} // namespace

Propeller::Propeller(std::function<std::chrono::microseconds()> clock)
  : _clock(clock)
{
  esp_err_t ret;
  spi_bus_config_t buscfg = {
    .mosi_io_num=MOSI,
    .miso_io_num=MISO,
    .sclk_io_num=CLK,
    .quadwp_io_num=-1,
    .quadhd_io_num=-1,
    .max_transfer_sz=0,
    .flags=SPICOMMON_BUSFLAG_NATIVE_PINS,
    .intr_flags=0
  };

  // The ESP has a very short CS-preamble
  // which means that the propeller doesn't have
  // enough time to prepare the output data
  // before we clock. Because of that the
  // CS is driven explicitly with GPIO.
  spi_device_interface_config_t devcfg = {
    .command_bits=0,
    .address_bits=0,
    .dummy_bits=0,
    .mode=0,  // CPOL=0, CPHA=0
    .duty_cycle_pos=0,
    .cs_ena_pretrans=0,
    .cs_ena_posttrans=0,
    .clock_speed_hz=SPI_SPEED,
    .input_delay_ns=0,
    .spics_io_num=-1,
    .flags=0,
    .queue_size=1,
    .pre_cb=nullptr,
    .post_cb=nullptr
  };
  //Initialize the SPI bus
  ret = spi_bus_initialize(HSPI_HOST, &buscfg, 0);
  ESP_ERROR_CHECK(ret);
  ret = spi_bus_add_device(HSPI_HOST, &devcfg, &_spi);
  ESP_ERROR_CHECK(ret);
  // This is our explicit CS driving setup
  gpio_config_t io_conf = {
    .pin_bit_mask = 1 << CS,
    .mode = GPIO_MODE_OUTPUT,
    .pull_up_en = gpio_pullup_t(0),
    .pull_down_en = gpio_pulldown_t(0),
    .intr_type = GPIO_INTR_DISABLE //disable interrupt
  };
  ret = gpio_config(&io_conf);
  ESP_ERROR_CHECK(ret);
  gpio_set_level(CS, 1);
  ESP_LOGI("prop", "spi initialized");
}

void Propeller::run(QueueHandle_t clock_queue)
{
  using namespace std::chrono_literals;

  auto playing = false;
  Sync24Beat beat;
  size_t peers = 0;
  double tempo;
  std::chrono::microseconds next_timepoint;


  // Because we drive CS ourselves, the propeller
  // captured an old timestamp and sits there
  // waiting to be clocked. This dummy-read
  // prevents us from providing a wrong timestamp
  propeller_transaction(_spi, std::array<uint8_t, 8>{0, 0, 0, 0, 0, 0, 0, 0});
  PropClock clock{_clock()};


  while(true)
  {
    ClockMessage message;
    auto suppress_valid_pulse = false;

    xQueueReceive(clock_queue, &message, portMAX_DELAY);
    std::visit(overloaded {
	[&](const NotPlaying& message) {
	  beat = message.beat;
	  tempo = message.tempo;
	  peers = message.peers;
	  suppress_valid_pulse = true;
	  next_timepoint = _clock() + 10000us;
	},
	[&](const StartPlaying& message) {
	  beat = message.beat;
	  tempo = message.tempo;
	  peers = message.peers;
	  assert(!playing);
	  ESP_LOGI("prop", "start playing");
	  playing = true;
	  next_timepoint = message.timepoint;
	},
	[&](const ClockAt& message) {
	  beat = message.beat;
	  tempo = message.tempo;
	  peers = message.peers;
	  next_timepoint = message.timepoint;
	},
	[&](const StopPlaying& message) {
	  beat = message.beat;
	  tempo = message.tempo;
	  peers = message.peers;
	  next_timepoint = message.timepoint;
	  assert(playing);
	  ESP_LOGI("prop", "stop playing");
	  playing = false;
	}
      }, message);

    auto clock_at_beat = clock.clock_for_timestamp(next_timepoint, playing);
    clock_at_beat.valid = !suppress_valid_pulse;
    auto clock_at_beat_u32 = clock_at_beat.to_u32();
    assert(!playing || (clock_at_beat_u32 & 2uL));
    uint32_t raw_clock_at_beat = htobe32(clock_at_beat_u32);
    ESP_LOGD("prop", "clock_at_beat: %x, raw: %x, valid: %i, playing: %i", clock_at_beat_u32, raw_clock_at_beat, is_valid(clock_at_beat_u32), is_playing(clock_at_beat_u32));
    // This rather weird construction was necessary due to
    // std::copy weirdly only copying the LSB.
    const uint8_t* pd = reinterpret_cast<const uint8_t*>(&raw_clock_at_beat);
    std::array<uint8_t, 8> output{*pd++, *pd++, *pd++, *pd++, 0, 0, 0, 0};
    ESP_LOG_BUFFER_HEX_LEVEL("prop", output.data(), output.size(), ESP_LOG_DEBUG);
    const auto then = _clock();
    const auto input = propeller_transaction(_spi, output);
    const auto now = _clock();

    clock.feed(input.clock_counter, (then + now) / 2 );
    ESP_LOGD("prop", "timestamp: %x, running: %lli, clock_speed: %f",  input.clock_counter, clock.running_clock, clock.clock_speed);
  }
}
